#include <pthread.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>

#ifndef XDIGITS
#define XDIGITS 8
#endif

struct chunk_header {
	size_t chunk_size;            //# of comitted bytes
	size_t requested_block_size;  //requested size
	size_t block_size;            //heap block size (rounded requested size)
	size_t block_offset;          //offset from start of block to (user-)pointer
	unsigned mgrclass;              //management class C/C++/C++ array
};
#define HEAPMGRCLASS_C          1
#define HEAPMGRCLASS_CPP        2
#define HEAPMGRCLASS_CPPARRAY   3

extern "C" void heap_panic(); //located in assembler file


#define MAXMEM ((size_t)(64*1024*1024))
#define PAGES (MAXMEM/PAGESIZE)
#define FILL_BYTE 0x7e

static unsigned in_use[PAGES] = {0};
static char *arena = 0;
static int fd_token[2] = {-1,-1};
static int fd_log = -1;
static size_t age_arm = 0;
#ifdef DELAYEDINIT
static bool pthread_self_safe = false;
#endif

extern "C" {
static void idh_reinitialize_after_fork();
}

static void idh_initialize() {
	arena = (char*)mmap(0, MAXMEM, PROT_NONE, MAP_PRIVATE|MAP_ANONYMOUS|MAP_NORESERVE, -1, 0);
	if(arena==(void*)MAP_FAILED)
		heap_panic();
	
	if(pipe(fd_token) != 0)
		heap_panic();
	char dummy;
	write(fd_token[1],&dummy,1);
	
	fd_log = open("heaplog",O_WRONLY|O_CREAT|O_TRUNC|O_APPEND,0644);
	
	//Note that this call has to be the after all other initialization
#ifndef DELAYEDINIT
	if(pthread_atfork(NULL,NULL,idh_reinitialize_after_fork) != 0)
		heap_panic();
#endif
}


extern "C" {
static void idh_reinitialize_after_fork() {
	//Single Unix Spec. guarantees that mapped memory etc. is inherited.
	//We could continue to use fd_token, but crashes in child processes
	//would affect the parent...
	//Note that we do not support Solaris fork(), only fork1() is supported
	
	//Create new token-pipe
	close(fd_token[0]);
	close(fd_token[1]);
	pipe(fd_token);
	char dummy;
	write(fd_token[1],&dummy,1);
	
	//Create a new log
	char logfilename[16];
	memcpy(logfilename,"heaplog.",8);
	int i = (int)getpid();
	char tmpbuf[5], *tmpbufptr=tmpbuf;
	while(i) {
		*tmpbufptr++ = (char)('0' + i%10);
		i/=10;
	}
	tmpbufptr--;
	char *p=logfilename+8;
	do {
		*p++ = *tmpbufptr;
	} while(tmpbufptr--!=tmpbuf);
	*p = '\0';
	
	close(fd_log);
	fd_log = open(logfilename,O_WRONLY|O_CREAT|O_TRUNC|O_APPEND,0644);
}
}

//Logging functions
//convert a 32/64-bit integer to a 8/16-digit hex string
static void i2x(size_t x, char *buf) {
	buf+=XDIGITS-1;
	for(int i=0; i<XDIGITS; i++) {
		static char xdigit[17]="0123456789abcdef";
		*(buf--) = xdigit[x&0xf];
		x>>=4;
	}
}

static void log_operationstart(int alloc, void *caller, int mgrclass) {
	char buf[XDIGITS+1];
#ifdef DELAYEDINIT
	if(pthread_self_safe)
		i2x(pthread_self(),buf);
	else
		i2x((size_t)-1,buf);
#else
	i2x(pthread_self(),buf);
#endif
	write(fd_log,buf,XDIGITS);
	write(fd_log," ",1);
	
	const char *op;
	if(alloc) {
		if(mgrclass==HEAPMGRCLASS_C)
			op = "malloc()";
		else if(mgrclass==HEAPMGRCLASS_CPP)
			op = "new     ";
		else /*if(mgrclass==HEAPMGRCLASS_CPPARRAY)*/
			op = "new[]   ";
	} else {
		if(mgrclass==HEAPMGRCLASS_C)
			op = "free()  ";
		else if(mgrclass==HEAPMGRCLASS_CPP)
			op = "delete  ";
		else /*if(mgrclass==HEAPMGRCLASS_CPPARRAY)*/
			op = "delete[]";
	}
	write(fd_log,op,8);
	write(fd_log," ",1);
	
	i2x((size_t)caller,buf);
	write(fd_log,buf,XDIGITS);
	write(fd_log," ",1);
}

static void log_range(size_t low, size_t high) {
	char buf[XDIGITS+1];
	i2x(low,buf);
	write(fd_log,buf,XDIGITS);
	write(fd_log,"-",1);
	i2x(high,buf);
	write(fd_log,buf,XDIGITS);
	write(fd_log," ",1);
}

static void log_norange() {
	write(fd_log,"                                  ",XDIGITS*2+2);
}

static void log_userptr(size_t ptr) {
	char buf[XDIGITS+1];
	i2x(ptr,buf);
	write(fd_log,buf,XDIGITS);
	write(fd_log," ",1);
}

static void log_size(size_t size) {
	char buf[XDIGITS+1];
	i2x(size,buf);
	write(fd_log,buf,XDIGITS);
}

static void log_nosize() {
	write(fd_log,"                ",XDIGITS);
}

static void log_operationfailure(const char *errmsg) {
	write(fd_log," ",1);
	write(fd_log,errmsg,strlen(errmsg));
	write(fd_log,"\n",1);
	close(fd_log);
	heap_panic();
}

static void log_operationsuccess()
{
	write(fd_log,"\n",1);
}




//Freemap management ----------------------------------------------------------

static void incremental_age_freemap() {
	if(in_use[age_arm]!=0 && in_use[age_arm]!=~0U)
		in_use[age_arm]--;
	age_arm = (age_arm+1)%PAGES;
}

static bool age_freemap() {
	unsigned oldest=~0U;
	for(size_t i=0; i<PAGES; i++)
		if(in_use[i]!=0 && in_use[i]<oldest)
			oldest = in_use[i];
	if(oldest==~0U) return false;
	for(size_t i=0; i<PAGES; i++)
		if(in_use[i]!=0 && in_use[i]!=~0U)
			in_use[i] -= oldest;
	return true;
}


static char *find_free_blocks(size_t blocks) {
	incremental_age_freemap();
	
	do {
		size_t start=0;
		size_t i=start;
		while(start<PAGES) {
			if(in_use[i])
				start = i = i+1;
			else {
				i++;
				if(i-start==blocks) {
					for(size_t j=start; j<i; j++)
						in_use[j] = ~0U;
					return arena+start*PAGESIZE;
				}
			}
		}
	} while(age_freemap());
	return 0;
}

static void free_blocks(void *p, size_t blocks) {
	size_t start = (((char*)p)-arena)/PAGESIZE;
	for(size_t i=0; i<blocks; i++)
		in_use[start+i] = (~0U)-1;
	incremental_age_freemap();
}


// Locking --------------------------------------------------------------------
static void idh_lock() {
	char dummy;
	while(read(fd_token[0],&dummy,1) !=1 )
		;
}

static void idh_unlock() {
	char dummy;
	while(write(fd_token[1],&dummy,1) != 1)
		;
}


static void *idh_allocmem(size_t requested_size, void *caller, int mgrclass) {
	if(!arena) idh_initialize();
	idh_lock();
	
	log_operationstart(1,caller,mgrclass);
	
#ifdef ALIGN
	size_t size;
	if(requested_size<3)
		size = requested_size;
	else if(requested_size<=5)
		size = (requested_size+3)&~3U;
	else if(requested_size<=8)
		size = (requested_size+7)&~7U;
	else
		size = (requested_size+15)&~15U;
#else
	size_t size = requested_size;
#endif
	//round up to nearest page size
	size_t chunk_size = (size+sizeof(chunk_header)+PAGESIZE-1)&~(PAGESIZE-1);
	size_t alloc_size = chunk_size+PAGESIZE;
	
	char *p = find_free_blocks(alloc_size/PAGESIZE);
	if(!p) {
		log_operationfailure("no free blocks left");
	}
	
	//change protection for all but the last page
	if(mprotect(p, chunk_size, PROT_READ|PROT_WRITE) !=0) {
		log_operationfailure("mprotect() failed");
	}
	
	//set memory to something
	memset(p, FILL_BYTE, chunk_size);
	
	//setup chunk housekeeping structure
	chunk_header *chp=(chunk_header *)p;
	chp->chunk_size   = chunk_size;
	chp->requested_block_size = requested_size;
	chp->block_size   = size;
	chp->block_offset = chp->chunk_size-chp->block_size;
	chp->mgrclass     = mgrclass;
	
	log_range((size_t)p,(size_t)p+chunk_size-1);
	log_userptr((size_t)p+chp->block_offset);
	log_size(requested_size);
	log_operationsuccess();
        
	idh_unlock();
	
	return p+chp->block_offset;
}

static inline chunk_header *ptr2chunk(void *p) {
//	return (chunk_header*) (((unsigned long)p)&~(PAGESIZE-1));
	return (chunk_header*) (((unsigned long)p-sizeof(chunk_header))&~(PAGESIZE-1));
}

static void idh_freemem(void *userptr, void *caller, unsigned mgrclass) {
#ifdef DELAYEDINIT
	static unsigned long ll=0;
	ll++;
	if(ll==DELAYEDINIT) {
		pthread_atfork(NULL,NULL,idh_reinitialize_after_fork);
		pthread_self_safe = true;
	}
#endif
	idh_lock();
	log_operationstart(0,caller,mgrclass);
	
	chunk_header *chp = ptr2chunk(userptr);
	void *chunkptr = (void*)chp;
	
	if((char*)userptr < arena || (char*)userptr>=arena+MAXMEM) {
		log_norange();
		log_userptr((size_t)userptr);
		log_nosize();
		log_operationfailure("Invalid user pointer");
	}
	
	size_t start_page = (((char*)userptr)-arena)/PAGESIZE;
	if(!in_use[start_page]) {
		log_norange();
		log_userptr((size_t)userptr);
		log_nosize();
		log_operationfailure("Invalid user pointer (probably freed)");
	}
	
	if(chp->block_offset != chp->chunk_size-chp->block_size) {
		log_norange();
		log_userptr((size_t)userptr);
		log_nosize();
		log_operationfailure("Bad pointer or header overwrite (1)");
	}
	//check for correct pointer
	if(userptr != (char*)chunkptr+chp->block_offset) {
		log_norange();
		log_userptr((size_t)userptr);
		log_nosize();
		log_operationfailure("Bad pointer or header overwrite (2)");
	}
	//check for overwrite of the unused bytes before the block
	for(unsigned char *checkp=((unsigned char*)chunkptr)+sizeof(chunk_header);
	    checkp<userptr;
	    checkp++)
	{
		if(*checkp != FILL_BYTE) {
			log_norange();
			log_userptr((size_t)userptr);
			log_nosize();
			log_operationfailure("Overwrite before user range");
		}
	}
	//check for incorrect mgrclass
	if(chp->mgrclass != mgrclass) {
		log_range((size_t)chunkptr,(size_t)chunkptr+chp->chunk_size-1);
		log_userptr((size_t)userptr);
		log_size(chp->requested_block_size);
		log_operationfailure("Mix of heap management classes");
	}
	
	//Ok
	size_t end = (size_t)chunkptr+chp->chunk_size-1;
	size_t requested_block_size = chp->requested_block_size;
	
	size_t blocks = chp->chunk_size/PAGESIZE;
	mprotect((char*)chunkptr, blocks*PAGESIZE, PROT_NONE);
	
	free_blocks(chp, blocks+1);
	
	log_range((size_t)chunkptr,end);
	log_userptr((size_t)userptr);
	log_size(requested_block_size);
	log_operationsuccess();
	
	idh_unlock();
}


static size_t user_size(void *p) {
	chunk_header *chp = ptr2chunk(p);
	return chp->block_size;
}

extern "C" void *idh_malloc(size_t size, void *caller) {
	return idh_allocmem(size,caller,HEAPMGRCLASS_C);
}
extern "C" void idh_free(void *ptr, void *caller) {
	if(ptr) idh_freemem(ptr,caller,HEAPMGRCLASS_C);
}

extern "C" void *idh_calloc(size_t nmemb, size_t size, void *caller) {
	void *ptr=idh_malloc(nmemb*size,caller);
	if(ptr) memset(ptr,0,nmemb*size);
	return ptr;
}

extern "C" void *idh_realloc(void *oldp, size_t size, void *caller) {
	void *newp=0;
	if(size>0) {
		if(!oldp)
			newp = idh_malloc(size,caller);
		else {
			newp = idh_malloc(size,caller);
			if(newp) {
				size_t oldsize = user_size(oldp);
				if(oldsize<size)
					memcpy(newp,oldp,oldsize);
				else
					memcpy(newp,oldp,size);
				idh_free(oldp,caller);
			}
		}
	} else {
		if(oldp)
			idh_free(oldp,caller);
		else
			;
	}
	return newp;
}

extern "C" void *idh_valloc(size_t size, void *caller) {
	size = (size+PAGESIZE-1)&~(PAGESIZE-1);
	return idh_allocmem(size,caller,HEAPMGRCLASS_C);
}

extern "C" char *idh_strdup(const char *s, void *caller) {
	size_t size = strlen(s)+1;
	char *p = (char*)idh_allocmem(size,caller,HEAPMGRCLASS_C);
	memcpy(p,s,size);
	return p;
}


extern "C" void *idh_new(size_t size, void *caller) {
	return idh_allocmem(size,caller,HEAPMGRCLASS_CPP);
}

extern "C" void idh_delete(void *p, void *caller) {
	if(p) idh_freemem(p,caller,HEAPMGRCLASS_CPP);
}

extern "C" void *idh_new_array(size_t size, void *caller) {
	return idh_allocmem(size,caller,HEAPMGRCLASS_CPPARRAY);
}

extern "C" void idh_delete_array(void *p, void *caller) {
	if(p) idh_freemem(p,caller,HEAPMGRCLASS_CPPARRAY);
}


#ifdef TEST1
int main() {
	char *p = (char*)idh_malloc(12,0);
	strcpy(p,"Hello world");                  //ok
	strcpy(p,"Hello my friend. What's your name?"); //error
	idh_free(p,0);
	return 0;
}

extern "C" void heap_panic() { }
#endif

#ifdef TEST2
int main() {
	char *p = (char*)malloc(12);
	strcpy(p,"Hello world");                  //ok
	strcpy(p,"Hello my friend. What's your name?"); //error
	free(p);
	return 0;
}
#endif

#ifdef TEST3
int main() {
	char *p = new char[12];
	strcpy(p,"Hello world");                  //ok
	strcpy(p,"Hello my friend. What's your name?"); //error
	delete[] p;
	return 0;
}
#endif

#ifdef TEST4
int main() {
	char *p = (char*)realloc(0,12);
	strcpy(p,"Hello world");                  //ok
	strcpy(p,"Hello my friend. What's your name?"); //error
	free(p);
	return 0;
}
#endif
